c     Reverted to F77: WJB 14 Aug 2003
c     Last change:  WJB  21 Dec 2000   11:42 am

       subroutine splinel2(x,y,n,gridpts,gridsize,degree,knots,p,
     &Sn,WORK1,WORK2,WORK3,Kxxo,xbs,predbs,betahat,fhat,Scopy,setmp,se)

c INPUT:
c x, y    = input regression data      (x's are sorted)
c n       = number of observations
c gridpts = locations of approximation evaluation
c gridsize= no. of gridpts
c degree  = degree of spline = order-1
c knots   = interior knots + 2 boundary knots
c p       = number of interior knots + order =
c                         # truncated power basis functions
c
c WORK AREA:
c Sn, WORK1, WORK2, WORK3, Kxxo 
c
c OUTPUT:
c fhat    = spline regression function evaluated at gridpts.

       implicit none
       INTEGER 	  gridsize, i, j, k, ell, n, degree, p, q
       INTEGER           WORK1(p)
       DOUBLE precision  x(n), y(n), gridpts(gridsize), predbs(n,p), 
     &  Sn(p,p+1), WORK2(p,p), WORK3(p+1), xo, Kxxo(n), zero, 
     &  betahat(p), knots(p-degree+1), fhat(gridsize), xbs(p), 
     &  meshsize, Snsum, yhat, Scopy(p,p+1),se(p),setmp(p)
       parameter         (zero=0.0d0)
       DOUBLE PRECISION  kernel

c  DLL_EXPORT        splinel2

       meshsize = (x(n) - x(1))/(gridsize-1)

       do i=1,gridsize
           gridpts(i) = x(1) + dfloat(i-1)*meshsize
       END do

       do i=1, n
        xo = x(i)
        q = p-degree+1
        call tspline(xo,degree,knots,q,p,xbs)
        do k=1,p
         predbs(i,k)=xbs(k)
        end do
       END do

       do k=1, p
        do ell =1, p
          Sn(ell,k) = zero
          do j=1,n
           Sn(ell,k) = Sn(ell,k) +  predbs(j,k)*predbs(j,ell)
          end do
        end do
       end do

       do k=1,p
        Snsum = zero
        do ell=1,p
         Snsum = Snsum + Sn(ell,k)
        end do
        if (ABS(Snsum) .LT. 10.0d0**(-10) ) then
         Sn(k,k) = 1.0d0
        end if
       end do

       do k=1, p
        Sn(k,p+1) = zero
        do j=1, n
         Sn(k,p+1) = Sn(k,p+1) + y(j)*predbs(j,k)
        end do
       end do

       do i=1,p
        do j=1,p+1
         Scopy(i,j)=Sn(i,j)
        end do
       end do

       call GAUSSPP(Scopy, p, betahat, WORK1, WORK2, WORK3)

c calculation of (un-normalized) standard errors (inverse of X^TX):

       do k=1, p

        do i=1,p
         do j=1,p
          Scopy(i,j)=Sn(i,j)
         end do
         if (i .eq. k) then
          Scopy(i,p+1)=1.0d0
         else
          Scopy(i,p+1)=0.0d0
         endif
        end do

        call GAUSSPP(Scopy, p, setmp, WORK1, WORK2, WORK3)

        se(k)=setmp(k)
       END do

      do i=1, gridsize
       xo = gridpts(i)
       call tspline(xo,degree,knots,q,p,xbs)
       fhat(i)=zero
       do k=1,p
        fhat(i) = fhat(i) + betahat(k)*xbs(k)
       end do
      END do

      return
      end subroutine

c!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

       subroutine tspline(x,degree,knots,q,p,xbs)

c knots = given knot sequence (including boundary knots)
c knotlength = number of interior knots + 2
c t = extended knot sequence (interior knots + order boundary knots on
c     each end)
c OUTPUT: xbs: p-array - bsplines evaluated at x
       implicit none

       INTEGER          degree, i, j, p, q, degree_tmp
       DOUBLE PRECISION knots(q), x, xbs(p)

c truncated basis functions:

       degree_tmp=degree
        DO i=2,q-1
         if (knots(i) .EQ. knots(i-1)) then    ! coalescent knots
          degree_tmp=degree_tmp-1
         else
          degree_tmp=degree                  ! distinct knot
         end if
         if (x .gt. knots(i)) then             ! truncated power
          xbs(i-1) = (x-knots(i))**degree_tmp
         else
          xbs(i-1) = 0.0d0
         endif
        END do

c polynomial basis functions:

        DO i=0,degree
         xbs(i+q-1) = x**i
       END DO

       return

       end subroutine

c     Reverted to F77: WJB 8 Aug 2003
c     Last change:  UOW  28 Jan 2001    5:03 pm

      subroutine locspline(x,y,n,gridpts,gridsize,h,degree,knots,p,Sn,
     &WORK1,WORK2,WORK3,Kxxo,xbs,predbs,betahat,fhat)

c   INPUT:
c   x, y    = input regression data      (x's are sorted)
c   n       = number of observations
c   gridpts = locations of approximation evaluation
c   gridsize= no. of gridpts
c   h       = bandwidth
c   hlow    = lower bound on bandwidth
c   degree  = degree of spline = order-1
c   knots   = interior knots + 2 boundary knots
c   p       = number of interior knots + order = # bspline basis functions
c 
c   WORK AREA:
c   Sn, WORK1, WORK2, WORK3, Kxxo 
c 
c   OUTPUT:
c   fhat    = kernel spline regression function evaluated at gridpts.

      implicit none
       INTEGER           gridsize, i, j, k, ell, n, degree, p, q
       INTEGER           WORK1(p)
       DOUBLE precision  h, x(n), y(n), gridpts(gridsize), predbs(n,p)
       DOUBLE precision                   Sn(p,p+1), WORK2(p,p)
       DOUBLE precision WORK3(p+1), xo, Kxxo(n), zero
       DOUBLE precision     betahat(p), knots(p-degree+1) 
       DOUBLE precision fhat(gridsize), xbs(p), meshsize, Snsum, yhat
       parameter         (zero=0.0d0)
       DOUBLE PRECISION  kernel

c DLL_EXPORT        locspline

       meshsize = (x(n) - x(1))/(gridsize-1)

c do i=1,gridsize
c     gridpts(i) = x(1) + dfloat(i-1)*meshsize
c END do

       do i=1, n
        xo = x(i)
        q = p-degree+1
        call bspline(xo,degree,knots,q,p,xbs)
        do k=1,p
         predbs(i,k)=xbs(k)
        end do
       END do

       do i=1, gridsize
        xo = gridpts(i)
         do j=1, n
          Kxxo(j) =  kernel((x(j)-xo)/h)   
c   this is the jth diagonal element of W
         end do
        do k=1, p
         do ell =1, p
           Sn(ell,k) = zero
           do j=1,n
            Sn(ell,k) = Sn(ell,k) + Kxxo(j)*predbs(j,k)*predbs(j,ell)
           end do
         end do
        end do

        do k=1,p
         Snsum = zero
         do ell=1,p
          Snsum = Snsum + Sn(ell,k)
         end do
         if (ABS(Snsum) .LT. 10.0d0**(-10) ) then
          Sn(k,k) = 1.0d0
         end if
        end do

        do k=1, p
         Sn(k,p+1) = zero
         do j=1, n
          Sn(k,p+1) = Sn(k,p+1) + Kxxo(j)*y(j)*predbs(j,k)
         end do
        end do

        call GAUSSPP(Sn, p, betahat, WORK1, WORK2, WORK3)
        call bspline(xo,degree,knots,q,p,xbs)

        fhat(i)=zero
        do k=1,p
         fhat(i) = fhat(i) + betahat(k)*xbs(k)
        end do

       END do

       return
       end subroutine

c !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

       subroutine bspline(x,degree,knots,q,p,xbs)

c   knots = given knot sequence (including boundary knots)
c   knotlength = number of interior knots + 2
c   t = extended knot sequence (interior knots + order boundary knots on
c       each end)
c   OUTPUT: xbs: p-array - bsplines evaluated at x
      implicit none

       INTEGER          degree, i, j, p, q, knotlength
       DOUBLE PRECISION knots(q),t(1000),tmp,tmp1,x
       DOUBLE PRECISION bvalue(1000,20),xbs(p)

c PRINT*, 'total no. of knots (boundary+interior): '
c READ*, knotlength
c   knotlength=3
c PRINT*, 'knots:'
c do i = 1, knotlength
c PRINT*, 'knot(',i,'):'
c READ*, knots(i)
c end do
c PRINT*, 'degree of spline: '
c READ*, degree
c PRINT*, 'evaluation point: '
c READ*, x

       knotlength = p + 1 - degree
c knots(1) = 1.0d0
c knots(2) = 5.5d0
c knots(3) = 10.0d0

       do i=1,degree+1
        t(i)=knots(1)
       end do

       do i=degree+2,degree+knotlength-1
        t(i) = knots(i-degree)
       end do

       do i=degree+knotlength, 2*degree+knotlength
        t(i) = knots(knotlength)
       end do

       do i=1,p+degree
        if ((x .ge. t(i)).AND.(x .LT. t(i+1))) then
         bvalue(i,1) = 1.0d0
        else
         bvalue(i,1) = 0.0d0
        endif
       end do

       if (degree .gt. 0) then

        do j = 1, degree

         do i=1,p+degree-j

          tmp = t(i+j+1)-t(i+1)
          if (tmp .gt. 0.0d0) then
           tmp =  (t(i+j+1)-x)/tmp*bvalue(i+1,j)
           if (tmp .le. 0.0d0) then
            tmp = 0.0d0
           end if
          else
           tmp = 0.0d0
          endif

          tmp1 = t(i+j)-t(i)
          if (tmp1 .gt. 0.0d0) then
           tmp1 = (x-t(i))/tmp1*bvalue(i,j)
           if (tmp1 .le. 0.0d0) then
            tmp1 = 0.0d0
           end if
          else
           tmp1=0.0d0
          end if

          bvalue(i,j+1) = tmp1 + tmp

         END do 
        END do
       END if

       DO i=1,p
        xbs(i) = bvalue(i,degree+1)
       END do

       return

       end subroutine

c     Smoother matrix diagonal computation: WJB 11 Aug 2003
c     Reverted to F77: WJB 8 Aug 2003
c     Last change:  UOW  28 Jan 2001    5:03 pm

      subroutine locsplinedf(x,y,n,h,degree,knots,p,Sn,Sny,
     &WORK1,WORK2,WORK3,Kxxo,xbs,predbs,betahat,fhat,hat)

c   INPUT:
c   x, y    = input regression data      (x's are sorted)
c   n       = number of observations
c   h       = bandwidth
c   hlow    = lower bound on bandwidth
c   degree  = degree of spline = order-1
c   knots   = interior knots + 2 boundary knots
c   p       = number of interior knots + order = # bspline basis functions
c 
c   WORK AREA:
c   Sn, Sny, WORK1, WORK2, WORK3, Kxxo 
c 
c   OUTPUT:
c   fhat    = kernel spline regression function evaluated at gridpts.
c   hat     = hat diagonal elements

      implicit none
       INTEGER           i, j, k, ell, n, degree, p, q
       INTEGER           WORK1(p)
       DOUBLE precision  h, x(n), y(n), predbs(n,p)
       DOUBLE precision  Sn(p,p+1), Sny(p,p+1), WORK2(p,p)
       DOUBLE precision WORK3(p+1), xo, Kxxo(n), zero
       DOUBLE precision     betahat(p), knots(p-degree+1) 
       DOUBLE precision fhat(n), hat(n), xbs(p), Snsum, yhat
       parameter         (zero=0.0d0)
       DOUBLE PRECISION  kernel

       do i=1, n
        xo = x(i)
        q = p-degree+1
        call bspline(xo,degree,knots,q,p,xbs)
        do k=1,p
         predbs(i,k)=xbs(k)
        end do
       END do

       do i=1, n
        xo = x(i)
         do j=1, n
          Kxxo(j) =  kernel((x(j)-xo)/h)   
c   this is the jth diagonal element of W
         end do
        do k=1, p
         do ell =1, p
           Sn(ell,k) = zero
           do j=1,n
            Sn(ell,k) = Sn(ell,k) + Kxxo(j)*predbs(j,k)*predbs(j,ell)
           end do
         end do
        end do

        do k=1,p
         Snsum = zero
         do ell=1,p
          Snsum = Snsum + Sn(ell,k)
         end do
         if (ABS(Snsum) .LT. 10.0d0**(-10) ) then
          Sn(k,k) = 1.0d0
         end if
        end do

        do k=1,p
        do j=1,p
         Sny(j,k)=Sn(j,k)
        end do
        end do

        do k=1, p
         Sn(k,p+1) = zero
         Sny(k,p+1)=zero
          Sn(k,p+1) = Sn(k,p+1) + Kxxo(i)*predbs(i,k)
         do j=1, n
          Sny(k,p+1) = Sny(k,p+1) + Kxxo(j)*y(j)*predbs(j,k)
         end do
        end do

        call GAUSSPP(Sny, p, betahat, WORK1, WORK2, WORK3)
        call bspline(xo,degree,knots,q,p,xbs)

        fhat(i)=zero
        do k=1,p
         fhat(i) = fhat(i) + betahat(k)*xbs(k)
        end do

        call GAUSSPP(Sn, p, betahat, WORK1, WORK2, WORK3)

        hat(i)=zero
        do k=1,p
         hat(i) = hat(i) + betahat(k)*xbs(k)
        end do

       END do

       return
       end subroutine

c !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

       DOUBLE PRECISION function kernel(x)

       implicit none
       DOUBLE PRECISION x, one, two, zero, fifteen, sixteen, pi
       parameter (one=1.0d0,zero=0.0d0,two=2.0d0,fifteen=15.0d0,
     & sixteen=16.0d0,pi=3.141593d0)
c if (ABS(x) .lt. one) then
c       kernel = fifteen*(one-x**2)**2/sixteen
c       else
c       kernel = zero
c end if

       kernel=exp(-x**2/two)/SQRT(two*pi)

       RETURN
       end

c !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

       SUBROUTINE GAUSSPP(A,N,X,NROW,M,TMP2)

c   Gaussian Elimination with Partial Pivoting to solve Ax = B;  B = A(.,N+1)
c   Algorithm 6.2 of Burden, Faires and Reynolds (1981)
c   SOLUTION = X(N)
c   work arrays: integer NROW(N), double M(N,N), double TMP2(N+1)

       IMPLICIT NONE
       INTEGER I, J, K, N, NCOPY, NROW(N), P
       DOUBLE PRECISION AMAX, A(N,N+1), M(N,N), TMP, X(N), TMP2(N+1)

       DO I = 1, N
        NROW(I) = I   
c   initialize row pointer
       END DO

       DO I = 1, N-1   
c   Elimination process
        AMAX = 0.0d0
        DO J = I, N
         IF (AMAX .LT. ABS(A(NROW(J),I))) THEN
         AMAX = ABS(A(NROW(J),I))
         P = J
         ENDIF
        END DO
c    IF (AMAX .EQ. 0.0d0) THEN
c       PRINT*, 'ERROR 1: NO UNIQUE SOLUTION EXISTS.'
c    END IF
c   ROW INTERCHANGE:
       NCOPY = NROW(I)
       NROW(I) = NROW(P)
       NROW(P) = NCOPY
       DO J = I + 1, N
        M(NROW(J),I) = A(NROW(J),I)/A(NROW(I),I)
        DO K = I+1, N+1
         TMP2(K) =  M(NROW(J),I)*A(NROW(I),K)
        END DO
        DO K = I+1, N+1
         A(NROW(J), K) = A(NROW(J), K) - TMP2(K)
        END DO
       END DO
      END DO
      IF (A(NROW(N),N) .EQ. 0.0d0) THEN
c     PRINT*, 'ERROR 2: NO UNIQUE SOLUTION EXISTS.'
c     CALL BREAK()
       A(NROW(N),N) = .0000001d0
      END IF
      X(N) = A(NROW(N),N+1)/A(NROW(N),N)
      DO I = N-1, 1, -1
       TMP = 0
       DO J = I+1, N
        TMP = A(NROW(I),J)*X(J)+TMP
       END DO
       X(I) = (A(NROW(I),N+1) - TMP)/A(NROW(I),I)
      END DO

      RETURN
      END SUBROUTINE

c !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


